#include "rfaclient/application.hpp"
#include "rfaclient/manager.hpp"
#include "rfaclient/system.hpp"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/format.hpp>
#include <boost/variant.hpp>

#include <iostream>
#include <vector>
#include <string>
#include <typeinfo>
#include <map>

#include <R.h>
#include <Rinternals.h>
#include <R_ext/Rdynload.h>

using namespace boost;
using namespace boost::posix_time;
using namespace rfaclient;
using namespace std;

#undef length

/* The main subscription function */
int init(int timeToRun);

/* Used to flag whether we need to re-read message dictionary */
static bool initialLoad = TRUE;	

/* The entrypoint back into R */
SEXP R_callback_function;

/* Store the configured items, RICs and FIDs in a map */
/* { (item,ric) => (f1, f2, ....fn) } */
typedef pair<string,string> serviceKey;
typedef map< serviceKey, vector<string*> > configurationMap;
configurationMap params;

/* Path to rfa config file */
string configPath;
string sessionName;
bool debugOn = false;

/* Ensure that the SEXP is a valid string type */
static bool checkString(SEXP var) {
	return (isString(var) && TYPEOF(STRING_ELT(var,0))==CHARSXP);
}

/**
* SEXP visitor that dispatches on the type of the 
* Boost variant
*/
class SEXP_visitor : public boost::static_visitor<SEXP>  
{
public:
	SEXP operator()(const int& i) const {
		return(ScalarInteger(i));
	}

	SEXP operator()(const long& l) const {
		return(ScalarReal(l));
	}

	SEXP operator()(const unsigned long& ul) const {
		return(ScalarReal(ul));
	}

	SEXP operator()(const double& d) const {
		return(ScalarReal(d));
	}

	SEXP operator()(const string& s) const {
		return mkString(s.c_str());
	}

	SEXP operator()(const bool& b) const {
		return(ScalarLogical(b));
	}

	// TODO
	SEXP operator()(const vector<char>& c) const { 
		//cout << "vector<char>" << endl;
		return NULL;
	}

	SEXP operator()(const NoData& nd) const { 
		return ScalarLogical(NA_LOGICAL);
	}

	SEXP operator()(const NoSupp& ns) const {
		return ScalarLogical(NA_LOGICAL); 
	}
};

/**
* Client Event Handler
*
*/
class RFAClient : public RfaClientBase {

	void
		doProcessEvent(const string& serviceName, const string& itemName,
		MarketDataMessageType msgType, const Fields& fields)
	{

		serviceKey key = make_pair(serviceName, itemName);

		// Do we have a mapping for this (service, item) pair?
		vector<string*>& fids = params[key];
		if (fids.empty()) {
			// Should never happen
			Rprintf("No configuration found for (%s,%s)\n", 
				serviceName.c_str(),
				itemName.c_str());
			return;
		}

		map<string,Field> callback_values;

		// Iterate over message fields and filter by our fids vector
		Fields::const_iterator it(fields.begin()), end(fields.end());
		for (; it != end; ++it) {

			vector<string*>::const_iterator myfids(fids.begin()), 
				myfids_end(fids.end());

			for (; myfids !=myfids_end; ++myfids) {
				if (strcmp(((string)it->first).c_str(), (*myfids)->c_str()) == 0) {
					callback_values[it->first] = it->second;
				}
			}
		}

		// Only call back to R if we actually have something to call back with
		if (callback_values.empty())
			return;

		// Add the item name (RIC) that just ticked to the data frame
		Field f = itemName;
		callback_values[string("ITEM")] = itemName;

		/*
		* Create a data frame to pass back to R
		*/
		SEXP df, names;
		int sz = callback_values.size();
		PROTECT(df = allocVector(VECSXP, sz));
		PROTECT(names = allocVector(STRSXP, sz));

		// Iterate thru the filtered field values
		map<string, Field>::const_iterator filtered_it(callback_values.begin()), 
			filtered_end(callback_values.end());
		int i = 0;
		for (; filtered_it != filtered_end; ++filtered_it) {
			SET_STRING_ELT(names, i, mkChar((filtered_it->first).c_str()));
			// Coerce the variant Field type to a SEXP
			SEXP tmp = boost::apply_visitor(SEXP_visitor(), filtered_it->second);
			SET_VECTOR_ELT(df, i, tmp);
			i++;
		}

		// Execute the R callback with the newly created data frame
		setAttrib(df, R_NamesSymbol, names);
		SETCADR(R_callback_function, df);
		eval(R_callback_function, R_GlobalEnv);
		UNPROTECT(2);
	}
	void
		doProcessEvent(const string& serviceName, const string& itemName,
		State state, StatusCode statusCode, const string& text)
	{
		; // Not interested in service events for now
	}
	void
		doProcessEvent(long contId, ContReplyState state,
		ContReplyStatusCode statusCode)
	{
		; // Not interested in connection events for now
	}

};


extern "C" {

	/**
	* R dyn.load entrypoint 
	* TODO Make this work
	*/
	__declspec(dllexport) void R_init_RFAClient(DllInfo* info) {}

	/**
	* Process a RIC/FID list
	* We want to process a list of vectors, like so:
	* [ 
	*	{ SRC_1, RIC_1, { Field_1...Field_n } } 
	*	{ SRC_2, RIC_2, { Field_1...Field_m } } 
	*  ...
	* ]
	*/
	SEXP process_arguments(SEXP args) {
		// Remove any cached configuration
		if (!params.empty())
			params.clear();

		for (int i = 0; i < Rf_length(args); ++i) {

			SEXP row = VECTOR_ELT(args, i);
			PrintValue(row);

			if (!isVector(row))
				_error("Must be a list of vectors");

			if (Rf_length(row) < 3)
				_error("Must supply at least service, item, and one field id");

			int j = 0;

			if (debugOn) {
				Rprintf("%s\n",(CHAR(STRING_ELT(row,j+1))));
				Rprintf("%s\n",(CHAR(STRING_ELT(row,j+2))));
			}

			string service(CHAR(STRING_ELT(row, j++)));
			string item(CHAR(STRING_ELT(row, j++)));
			serviceKey key = make_pair(service,item);

			vector<string*> fids;
			for (; j < Rf_length(row); j++) {
				SEXP entry;
				PROTECT(entry = STRING_ELT(row, j));
				string* str = new string(CHAR(entry));
				fids.push_back(str);
				UNPROTECT(1);
			}

			params[key] = fids;
		}

		return ScalarInteger(0);
	}


	/**
	* The entrypoint from R code
	* Arguments: (timeout, interestSpecification, callbackFunction)
	*
	*/
	__declspec(dllexport) SEXP subscribe(SEXP t, SEXP spec, SEXP callback, SEXP session, SEXP config, SEXP debug) {
		int* time = INTEGER(t);

		if (!isVector(spec))
			Rf_error("Specification must be a list");

		if (!isFunction(callback))
			Rf_error("'callback' must be a function");

		if (!checkString(session)) 
			Rf_error("session must be defined");

		if (!isLogical(debug))
			Rf_error("invalid debug parameter");
		
		if (*LOGICAL(debug))
			debugOn = true;

		sessionName = CHAR(STRING_ELT(session,0));
		if (debugOn)
			Rprintf("using sessionName = %s\n", sessionName.c_str());

		if (!checkString(config)) 
			Rf_error("config path must be defined");

		configPath = CHAR(STRING_ELT(config,0));
		if(debugOn)
			Rprintf("configPath = %s\n", configPath.c_str());
		
		// Initialize the configuration map
		process_arguments(spec);

		// Initialize the callback object
		PROTECT(R_callback_function = lang2(callback, R_NilValue));

		int ret = init(*time);
		UNPROTECT(1);
		return ScalarReal(ret);
	}

} // end extern "C"

/**
* Initialize the underlying RFA machinery
* enumtype.def and appendix_a and a config file called rfa.cfg
* are assumed to be in the current working directory.
*/
int init(int timeToRun)
{
	try {
		RfaApplication rfa(configPath.c_str(), 
			"appendix_a", 
			"enumtype.def",
			initialLoad);
		
		RfaManager* manager;
		RFAClient client;
		initialLoad = FALSE;	

		string connectionNameKey("Sessions\\");
		connectionNameKey.append(sessionName).append("\\connectionList");
		const string connection = rfa.getConfigValue(connectionNameKey);
		
		if (debugOn) 
			Rprintf("Value from config database for key [%s]=>[%s]\n", connectionNameKey.c_str(), connection.c_str());

		if (connection.empty()) {
			Rf_error("No connection defined for key %s\n", connectionNameKey.c_str());
		}

		// Locate a username given a connection name
		// Assume \Connections\<connection name>\UserName
		string usernameKey("Connections\\");
		usernameKey.append(connection).append("\\UserName");
		const string username = rfa.getConfigValue(usernameKey);

		if (debugOn) 
			Rprintf("Value from config database for key [%s]=>[%s]\n", usernameKey.c_str(), username.c_str());

		if (!username.empty()) {
			Buffer buf((const unsigned char*)username.c_str(), 
				username.length(), username.length());
			TokenizedPrincipalIdentity identity;
			identity.setBuffer(buf);
			manager = new RfaManager(sessionName.c_str(), client, 
				static_cast<PrincipalIdentity*>(&identity), 1000, 0);
		}
		else {
			manager = new RfaManager(sessionName.c_str(), client, 1000);
		}
		
		/* Subscribe to the RICs */
		configurationMap::const_iterator it(params.begin()), end(params.end());
		for (; it != end; ++it) 
			manager->subscribe((it->first).first, (it->first).second);

		// Put this thread to sleep for the specified time
		// (Updates happen on the RFA thread)
		msleep((long)timeToRun);

		it = params.begin();
		for (; it != end; ++it) {
			manager->unsubscribe((it->first).first, 
				(it->first).second);
		}

		manager->clear();
		delete manager;
		return 0;

	} catch (const rfa::common::Exception& e) {
		cerr << e.getClassification() << 
			e.getErrorType() << e.getServerity() << endl;
	} catch (const exception& e) {
		cerr << e.what() << endl;
	}
	return 1;
}



