#define COMETRFAPP_BUILD
#include "basicpublisher.hpp"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/format.hpp>

using namespace boost;
using namespace boost::posix_time;
using namespace rfaclient;
using namespace std;

void
RfaBasicPublisher::publish(const string& itemName, const FieldsPtr& fields,
                           bool image)
{
    // Lock must be held.

    if (!publish_)
        return;

    // If not subscribed, subscribe now.

    Items::iterator it(items_.find(itemName));
    if (it == items_.end()) {

        it = items_.insert(make_pair(itemName,
                                     make_pair(queue<FieldsPtr>(), false)))
            .first;
        manager_->subscribe(serviceName_, itemName);
    }

    // Add fields to queue.

    if (fields) {
        it->second.first.push(fields);
        ++pending_;
    }

    if (image)
        it->second.second = true;
    else if (!it->second.second)
        return; // No image.

    while (!it->second.first.empty()) {
        FieldsPtr fields(it->second.first.front());
        it->second.first.pop();
        --pending_;
        contIds_.insert(manager_->publish(serviceName_, itemName, *fields));
    }

    // Destructor may be waiting for publications to complete.

    cond_.notify_all();
}

void
RfaBasicPublisher::doProcessEvent(const string& serviceName,
                                  const string& itemName,
                                  MarketDataMessageType msgType,
                                  const Fields& fields)
{
    writeLog(LOGINFO, SUBMARKETDATA, str
             (format("service=[%1%], item=[%2%], msgtype=[%3%]")
              % serviceName % itemName % toString(msgType)), logger_);

    if (rfa::sessionLayer::MarketDataItemEvent::Image == msgType) {

        // Now that initial snapshot has been received, publish any queue
        // fields.

        mutex::scoped_lock lock(mutex_);
        publish(itemName, FieldsPtr(), true);
    }
}

void
RfaBasicPublisher::doProcessEvent(const string& serviceName,
                                  const string& itemName, State state,
                                  StatusCode statusCode, const string& text)
{
    using namespace rfa::sessionLayer;

    LogLevel level((state == MarketDataItemStatus::Ok
                    || state == MarketDataItemStatus::NoChange)
                   ? LOGDEBUG0 : LOGERROR);

    writeLog(level, SUBMARKETDATA, str
             (format("service=[%1%], item=[%2%], state=[%3%],"
                     " status_code=[%4%], text=[%5%]")
              % serviceName % itemName % state % statusCode % text),
             logger_);
}

void
RfaBasicPublisher::doProcessEvent(long contId, ContReplyState state,
                                  ContReplyStatusCode statusCode)
{
    using namespace rfa::sessionLayer;

    if (state != MarketDataItemContReplyStatus::Ack)
        writeLog(LOGERROR, SUBMARKETDATA, str
                 (format("information could not be published to DTS:"
                         " state=[%1%], status_code=[%2%]")
                  % state % statusCode), logger_);

    mutex::scoped_lock lock(mutex_);
    contIds_.erase(contId);
    cond_.notify_all();
}

RfaBasicPublisher::~RfaBasicPublisher() NOTHROW
{
    {
        xtime xt;
        xtime_get(&xt, boost::TIME_UTC);
        xt.sec += timeout_; // Seconds to wait for subscription to complete.

        mutex::scoped_lock lock(mutex_);
        while (pending_ || !contIds_.empty())
            if (!cond_.timed_wait(lock, xt)) {

                // Timeout.

                writeLog(LOGWARN, SUBMARKETDATA,
                         "timeout waiting for DTS publication, giving-up");
                break;
            }
        publish_ = false;
    }
    manager_.reset();
}

RfaBasicPublisher::RfaBasicPublisher(const string& serviceName,
                                     unsigned timeout,
                                     const LoggerPtr& logger)
    : serviceName_(serviceName),
      timeout_(timeout),
      logger_(logger),
      pending_(0),
      publish_(true)
{
    manager_.reset(new RfaManager("COMETSession", *this, 1000, logger));
}

void
RfaBasicPublisher::prepare(const string& itemName)
{
    // Subscribe if not already subscribed.

    Items::const_iterator it(items_.find(itemName));
    if (it == items_.end()) {
        items_.insert(make_pair(itemName,
                                make_pair(queue<FieldsPtr>(), false)));
        manager_->subscribe(serviceName_, itemName);
    }
}

void
RfaBasicPublisher::publish(const string& itemName, const FieldsPtr& fields)
{
    mutex::scoped_lock lock(mutex_);
    publish(itemName, fields, false);
}

bool
RfaBasicPublisher::empty() const
{
    mutex::scoped_lock lock(mutex_);
    return !pending_;
}
