#define COMETRFAPP_BUILD
#include "positionpublisher.hpp"

#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/format.hpp>

using namespace boost;
using namespace boost::posix_time;
using namespace rfaclient;
using namespace std;

namespace {

    Fields
    toFields(const vector<pair<string, double> >& positions,
             const RfaPositionSettings& settings, bool zero)
    {
        string symbols;
        Fields fields;

        vector<pair<string, double> >::const_iterator it(positions.begin()),
            end(positions.end());
        for (unsigned i(1); it != end; ++it, ++i) {

            // Append symbol to list.

            if (1 < i)
                symbols += ',';
            symbols += it->first;

            // Add initial position for symbol.

            fields[str(format(settings.fieldTemplate_) % i)]
                = zero ? 0.0 : it->second;
        }

        fields[settings.symbolsField_] = symbols;
        fields[settings.timestampField_] = to_iso_string
            (microsec_clock::universal_time());

        return fields;
    }

    unsigned
    symbolIndex(const vector<pair<string, double> >& positions,
                const string& symbol)
    {
        vector<pair<string, double> >::const_iterator it(positions.begin()),
            end(positions.end());
        for (unsigned i(1); it != end; ++it, ++i)
            if (it->first == symbol)
                return i;

        throw MarketDataError(str(format("invalid position symbol '%1%'")
                                  % symbol));
    }
}

 void
RfaPositionPublisher::publish()
{
    // This function assumes that a lock is already in place.

    positions_[settings_.statusField_]
        = plRates_[settings_.statusField_]
        = string(open_ ? "OPEN" : "CLOSED");

    if (publish_) {
        manager_->publish(settings_.serviceName_, settings_.positionName_,
                          positions_);
        manager_->publish(settings_.serviceName_, settings_.plRateName_,
                          plRates_);
    }
}

 void
RfaPositionPublisher::doProcessEvent(const string& serviceName,
                                     const string& itemName,
                                     MarketDataMessageType msgType,
                                     const Fields& fields)
{
    writeLog(LOGINFO, SUBPOSITION, str
             (format("service=[%1%], item=[%2%], msgtype=[%3%]")
              % serviceName % itemName % toString(msgType)), logger_);

    if (rfa::sessionLayer::MarketDataItemEvent::Image == msgType) {

        // Now that initial snapshot has been received, publish position.

        mutex::scoped_lock lock(mutex_);
        publish();
    }
}

 void
RfaPositionPublisher::doProcessEvent(const string& serviceName,
                                     const string& itemName, State state,
                                     StatusCode statusCode,
                                     const string& text)
{
    using namespace rfa::sessionLayer;

    LogLevel level((state == MarketDataItemStatus::Ok
                    || state == MarketDataItemStatus::NoChange)
                   ? LOGDEBUG0 : LOGERROR);

    writeLog(level, SUBPOSITION, str
             (format("service=[%1%], item=[%2%], state=[%3%],"
                     " status_code=[%4%], text=[%5%]")
              % serviceName % itemName % state % statusCode % text), logger_);
}

 void
RfaPositionPublisher::doProcessEvent(long contId, ContReplyState state,
                                     ContReplyStatusCode statusCode)
{
    using namespace rfa::sessionLayer;

    if (state == MarketDataItemContReplyStatus::Ack) {

        // Checkpoint: positions published successfully.

        checkpoint(SUBPOSITION);

    } else
        writeLog(LOGERROR, SUBPOSITION, str
                 (format("information could not be published to DTS:"
                         " state=[%1%], status_code=[%2%]")
                  % state % statusCode), logger_);
}


RfaPositionPublisher::~RfaPositionPublisher() NOTHROW
{
    {
        mutex::scoped_lock lock(mutex_);
        publish_ = false;
    }
    manager_.reset();
}


RfaPositionPublisher
::RfaPositionPublisher(const vector<pair<string, double> >& initial,
                       const RfaPositionSettings& settings,
                       const LoggerPtr& logger)
    : initial_(initial),
      settings_(settings),
      logger_(logger),
      positions_(toFields(initial, settings, false)),
      plRates_(toFields(initial, settings, true)),
      open_(false),
      publish_(true)
{
    manager_.reset(new RfaManager(settings.sessionName_, *this,
                                  settings.dispatchTimeout_,
                                  bind(&RfaPositionPublisher::republish,
                                       this), logger));
    manager_->subscribe(settings_.serviceName_, settings.positionName_);
    manager_->subscribe(settings_.serviceName_, settings.plRateName_);
}

 void
RfaPositionPublisher::putPosition(const string& currency, double value)
{
    writeLog(LOGINFO, SUBPOSITION, str
             (format("updating position: currency=[%1%], value=[%2$4.2f]")
              % currency % value), logger_);

    mutex::scoped_lock lock(mutex_);
    positions_[settings_.timestampField_] = to_iso_string
        (microsec_clock::universal_time());

    unsigned i(symbolIndex(initial_, currency));
    string fieldName(str(format(settings_.fieldTemplate_) % i));
    positions_[fieldName] = value;
    publish();
}

 void
RfaPositionPublisher::putPositions(const vector<
                                   pair<string, double> >& positions)
{
    mutex::scoped_lock lock(mutex_);
    positions_[settings_.timestampField_] = to_iso_string
        (microsec_clock::universal_time());

    vector<pair<string, double> >::const_iterator it(positions.begin()),
        end(positions.end());
    for (; it != end; ++it) {

        writeLog(LOGINFO, SUBPOSITION, str
                 (format("updating position: currency=[%1%], value=[%2$4.2f]")
                  % it->first % it->second), logger_);

        unsigned i(symbolIndex(initial_, it->first));
        string fieldName(str(format(settings_.fieldTemplate_) % i));
        positions_[fieldName] = it->second;
    }
    writeLog(LOGINFO, SUBPOSITION, "publishing batch update", logger_);
    publish();
}

void
RfaPositionPublisher::putPlRate(const string& currency, double value)
{
    writeLog(LOGINFO, SUBPOSITION, str
             (format("updating position: currency=[%1%], value=[%2$4.2f]")
              % currency % value), logger_);

    mutex::scoped_lock lock(mutex_);
    plRates_[settings_.timestampField_] = to_iso_string
        (microsec_clock::universal_time());

    unsigned i(symbolIndex(initial_, currency));
    string fieldName(str(format(settings_.fieldTemplate_) % i));
    plRates_[fieldName] = value;
    publish();
}

 void
RfaPositionPublisher::putPlRates(const vector<pair<string, double> >& plRates)
{
    mutex::scoped_lock lock(mutex_);
    plRates_[settings_.timestampField_] = to_iso_string
        (microsec_clock::universal_time());

    vector<pair<string, double> >::const_iterator it(plRates.begin()),
        end(plRates.end());
    for (; it != end; ++it) {

        writeLog(LOGINFO, SUBPOSITION, str
                 (format("updating plrate: currency=[%1%], value=[%2$8.6f]")
                  % it->first % it->second), logger_);

        unsigned i(symbolIndex(initial_, it->first));
        string fieldName(str(format(settings_.fieldTemplate_) % i));
        plRates_[fieldName] = it->second;
    }
    writeLog(LOGINFO, SUBPOSITION, "publishing batch update", logger_);
    publish();
}

 void
RfaPositionPublisher::republish()
{
    writeLog(LOGDEBUG0, SUBPOSITION, "republishing positions", logger_);
    mutex::scoped_lock lock(mutex_);
    publish();
}

 void
RfaPositionPublisher::setOpen(bool open)
{
    mutex::scoped_lock lock(mutex_);
    positions_[settings_.timestampField_]
        = plRates_[settings_.timestampField_]
        = to_iso_string(microsec_clock::universal_time());

    open_ = open;
    publish();
}
