#define COMETRFAPP_BUILD
#include "rateprovider.hpp"

#include "manager.hpp"

#include "rateforwarder.hpp"

#include <boost/format.hpp>
#include <boost/date_time/gregorian/gregorian.hpp>
#include <boost/optional.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/thread.hpp>
#include <boost/tuple/tuple.hpp>

using namespace boost;
using namespace boost::gregorian;
using namespace rfaclient;
using namespace std;

namespace {

    typedef tuple<string, optional<double>, optional<double>,
                  optional<date> > RfaRate;

    bool
    isFuture(const date& d)
    {
        date now(boost::gregorian::day_clock::universal_day());
        return !d.is_not_a_date() && now < d;
    }

    void
    unpack(const string& serviceName, const string& itemName, RfaRate& rate,
           const Fields& fields, const RfaRateSettings& settings)
    {
        Fields::const_iterator it;

        // Bid.

        if ((it = fields.find(settings.bidField_)) != fields.end())
            get<1>(rate) = get<Real>(it->second);

        // Ask.

        if ((it = fields.find(settings.askField_)) != fields.end())
            get<2>(rate) = get<Real>(it->second);

        // ValueDate.

        if ((it = fields.find(settings.valueDateField_)) != fields.end()) {

            // Now convert from MS OADate field.  OADates are the number of
            // days from 30 Dec 1899.

            static const date epoch(1899, 12, 30);

            long d(static_cast<long>(get<Real>(it->second)));
            if (d <= 0)
                throw MarketDataError
                    (str(format("invalid value date: service=[%1%],"
                                " item=[%2%], valuedate=[%3%]")
                         % serviceName % itemName % d));

            date valdate(epoch + date_duration(d));
            if (!isFuture(valdate))
                throw MarketDataError
                    (str(format("value date not in future: service=[%1%],"
                                " item=[%2%], valuedate=[%3%]")
                         % serviceName % itemName % valdate));

            get<3>(rate) = valdate;
        }
    }

    class RateProvider : public RateProviderBase, public RfaClientBase {

        const RfaRateSettings settings_;
        LoggerPtr logger_;
        mutable mutex mutex_;
        RateForwarder forwarder_;
        scoped_ptr<RfaManager> manager_;
        map<string, RfaRate> rates_;

        // RateCacheBase.

        bool
        doLastRate(const std::string& symbol, Rate& rate) const
        {
            mutex::scoped_lock lock(mutex_);
            return forwarder_.lastRate(symbol, rate);
        }

        // RateProviderBase.

        void
        doSubscribe(const string& symbol, RateHandlerBase& handler)
        {
            mutex::scoped_lock lock(mutex_);
            forwarder_.subscribe(symbol, handler);

            string itemName(str(format(settings_.itemTemplate_) % symbol));
            if (rates_.find(itemName) == rates_.end()) {

                // This is a new symbol so subscribe to DTS.

                manager_->subscribe(settings_.serviceName_, itemName);
                RfaRate rate;
                get<0>(rate) = symbol;
                rates_.insert(make_pair(itemName, rate));
            }
        }

        void
        doUnSubscribe(const string& symbol, RateHandlerBase& handler)
        {
            // Note: remain subscribed to DTS.

            mutex::scoped_lock lock(mutex_);
            forwarder_.unSubscribe(symbol, handler);
        }

        void
        doUnSubscribe(RateHandlerBase& handler)
        {
            mutex::scoped_lock lock(mutex_);
            forwarder_.unSubscribe(handler);
        }

        void
        doUnSubscribe()
        {
            mutex::scoped_lock lock(mutex_);
            forwarder_.unSubscribe();
        }

        // RfaClientBase.

        void
        doProcessEvent(const string& serviceName, const string& itemName,
                       MarketDataMessageType msgType, const Fields& fields)
        {
            mutex::scoped_lock lock(mutex_);

            if (serviceName != settings_.serviceName_)
                throw MarketDataError(str(format("invalid service '%1%'")
                                          % serviceName));

            map<string, RfaRate>::iterator it(rates_.find(itemName));
            if (it == rates_.end())
                throw MarketDataError
                    (str(format("no subscription to service '%1%',"
                                " item '%2%'") % serviceName % itemName));

            // Unpack rate data into symbol's current image.

            unpack(serviceName, itemName, it->second, fields, settings_);

            string symbol;
            optional<double> bid;
            optional<double> ask;
            optional<date> valueDate;
            tie(symbol, bid, ask, valueDate) = it->second;

            // TODO: given that an initial image is sent from DTS, is it
            // possible, under normal circumstances, to have an incomplete
            // image?

            writeLog(LOGDEBUG0, SUBMARKETDATA, str
                     (format("service=[%1%], item=[%2%], symbol=[%3%],"
                             " bid=[%4%], ask=[%5%], valdate=[%6%]")
                      % serviceName % itemName % symbol % bid % ask
                      % valueDate), logger_);

            // Only advise if both bid and ask have been set.

            if (!!bid && !!ask) {
                date tradeDate(day_clock::universal_day());
                forwarder_.rate(symbol, *bid, *ask, tradeDate, valueDate);
            }
        }
        void
        doProcessEvent(const string& serviceName, const string& itemName,
                       State state, StatusCode statusCode, const string& text)
        {
            using namespace rfa::sessionLayer;

            mutex::scoped_lock lock(mutex_);
            LogLevel level((state == MarketDataItemStatus::Ok
                            || state == MarketDataItemStatus::NoChange)
                           ? LOGINFO : LOGERROR);

            writeLog(level, SUBMARKETDATA, str
                     (format("service=[%1%], item=[%2%], state=[%3%],"
                             " status_code=[%4%], text=[%5%]")
                      % serviceName % itemName % state % statusCode % text),
                     logger_);
        }
        void
        doProcessEvent(long contId, ContReplyState state,
                       ContReplyStatusCode statusCode)
        {
            using namespace rfa::sessionLayer;

            mutex::scoped_lock lock(mutex_);

            if (state != MarketDataItemContReplyStatus::Ack)
                writeLog(LOGERROR, SUBMARKETDATA, str
                         (format("information could not be published to DTS:"
                                 " state=[%1%], status_code=[%2%]")
                          % state % statusCode), logger_);
        }

    public:
        ~RateProvider() NOTHROW
        {
            manager_.reset();
        }
        RateProvider(const RfaRateSettings& settings, const LoggerPtr& logger)
            : settings_(settings),
              logger_(logger)
        {
            manager_.reset(new RfaManager(settings.sessionName_, *this,
                                          settings.dispatchTimeout_, logger));
        }
    };
}

 RateProviderPtr
comet::createRfaRateProvider(const RfaRateSettings& settings,
                             const LoggerPtr& logger)
{
    return RateProviderPtr(new RateProvider(settings, logger));
}
