#define rfaclientRFA_BUILD
#include "utility.hpp"

#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>

#include <vector>

using namespace boost;
using namespace rfaclient;
using namespace boost;
using namespace std;

namespace {

    template <typename T>
    bool
    convert(Field& dst, TibField& src)
    {
		
        if (0 == src.Size()) {
            dst = T();
            return false;
        }

        T v;
        if (src.Convert(v) != TIBMSG_OK)
            throw MarketDataError("TibField::Convert() failed");
        dst = v;
        return true;
    }

    template <>
    bool
    convert<string>(Field& dst, TibField& src)
    {
		
        if (0 == src.Size()) {
            dst = ""; // TODO why does dst=string(); cause 'bad allocation'?
            return false;
        }

        vector<char> v(src.Size());

        if (src.Convert(&v[0], v.size()) != TIBMSG_OK)
            throw MarketDataError("TibField::Convert() failed");

        // Remove null if one exists.

        if ('\0' == v.back())
            v.pop_back();

        dst = string(v.begin(), v.end());
        return true;
    }

    template <>
    bool
    convert<bool>(Field& dst, TibField& src)
    {
        if (0 == src.Size()) {
            dst = false;
            return false;
        }

        long v;
        if (src.Convert(v) != TIBMSG_OK)
            throw MarketDataError("TibField::Convert() failed");
        dst = v ? true : false;
        return true;
    }

    template <>
    bool
    convert<vector<char> >(Field& dst, TibField& src)
    {
        vector<char> v(src.Size());
        if (0 == src.Size()) {
            dst = v;
            return false;
        }

        if (src.Convert(&v[0], v.size()) != TIBMSG_OK)
            throw MarketDataError("TibField::Convert() failed");
        dst = v;
        return true;
    }
}

 bool
rfaclient::unpack(Field& dst, TibField& src)
{
    bool valid(false);
    switch (src.Type()) {
    case TIBMSG_NODATA:
        dst = NoData();
        break;
    case TIBMSG_MESSAGE:
        dst = NoSupp();
        break;
    case TIBMSG_STRING:
        valid = convert<string>(dst, src);
        break;
    case TIBMSG_OPAQUE:
        dst = NoSupp();
        break;
    case TIBMSG_BOOLEAN:
        valid = convert<bool>(dst, src);
        break;
    case TIBMSG_INT:
        valid = convert<long>(dst, src);
        break;
    case TIBMSG_UINT:
        valid = convert<unsigned long>(dst, src);
        break;
    case TIBMSG_REAL:
        valid = convert<double>(dst, src);
        break;
    case TIBMSG_ARRAY:
        valid = convert<vector<char> >(dst, src);
        break;
    case TIBMSG_PARTIAL:
        dst = NoSupp();
        break;
    case TIBMSG_IPDATA:
        dst = NoSupp();
        break;
    case TIBMSG_ANSIPAGE:
        dst = NoSupp();
        break;
    case TIBMSG_MAXVALID:
        assert(0);
        break;
    }
    return valid;
}

 const char*
rfaclient::toString(MarketDataMessageType msgType)
{
    const char* s;
    switch (msgType) {
    case rfa::sessionLayer::MarketDataItemEvent::Update:
        s = "Update";
        break;
    case rfa::sessionLayer::MarketDataItemEvent::Image:
        s = "Image";
        break;
    case rfa::sessionLayer::MarketDataItemEvent::UnsolicitedImage:
        s = "UnsolicitedImage";
        break;
    case rfa::sessionLayer::MarketDataItemEvent::Correction:
        s = "Correction";
        break;
    case rfa::sessionLayer::MarketDataItemEvent::ClosingRun:
        s = "ClosingRun";
        break;
    case rfa::sessionLayer::MarketDataItemEvent::Status:
        s = "Status";
        break;
    case rfa::sessionLayer::MarketDataItemEvent::Rename:
        s = "Rename";
        break;
    case rfa::sessionLayer::MarketDataItemEvent::PermissionData:
        s = "PermissionData";
        break;
    default:
        s = "Unknown";
        break;
    }
    return s;
}

 string
rfaclient::toString(const rfa::common::Exception& e)
{
	string severity;
	switch (e.getServerity()) {
    case rfa::common::Exception::Error:
        severity = "Error";
        break;
    case rfa::common::Exception::Warning:
        severity = "Warning";
        break;
    case rfa::common::Exception::Information:
        severity = "Information";
        break;
    default:
        severity = "Unknown";
        break;
	}

	string classification;
	switch (e.getClassification()) {
    case rfa::common::Exception::Anticipated:
        classification = "Anticipated";
        break;
    case rfa::common::Exception::Internal:
        classification = "Internal";
        break;
    case rfa::common::Exception::External:
        classification = "External";
        break;
    case rfa::common::Exception::IncorrectAPIUsage:
        classification = "IncorrectAPIUsage";
        break;
    case rfa::common::Exception::ConfigurationError:
        classification = "ConfigurationError";
        break;
    default:
        classification = "Unknown";
        break;
	}

	string type, status, details;
	switch (e.getErrorType()) {
    case rfa::common::Exception::InvalidUsageException:
        {
			type = "InvalidUsageException";
			const rfa::common::InvalidUsageException& actual
                (static_cast<const rfa::common::InvalidUsageException&>(e));
			status = actual.getStatus().getStatusText();
			break;
        }
    case rfa::common::Exception::InvalidConfigurationException:
        {
			type = "InvalidConfigurationException";
			const rfa::common::InvalidConfigurationException& actual
                (static_cast<const rfa::common
                 ::InvalidConfigurationException&>(e));
            status = actual.getStatus().getStatusText();
            details = string(actual.getParameterName()).append("=")
                .append(actual.getParameterValue());
			break;
        }
#if defined(_WIN32X) // Only valid for RFA6+
    case rfa::common::Exception::SystemException:
        {
			type = "SystemException";
			const rfa::common::SystemException& actual
                (static_cast<const rfa::common::SystemException&>(e));
			status = actual.getStatus().getStatusText();
			break;
        }
#endif // _WIN32
    default:
        status = "Unknown";
        break;
	}

    return str(format("rfa exception: severity='%1%', class='%2%',"
                      " type='%3%', status='%4%', details='%5%'")
               % severity
               % classification
               % type
               % status
               % details);
}

 pair<string, string>
rfaclient::splitPair(const string& s, const char* delims)
{
    // Split into name/value pair.

    string::size_type i(s.find_first_of(delims));
    if (string::npos == i)
		throw std::runtime_error(string("invalid line: ").append(s));

    string name(trim_right_copy(s.substr(0, i)));
    if (name.empty())
		throw std::runtime_error("empty field name");

    return make_pair(name, trim_left_copy(s.substr(i + 1)));
}


 std::string
rfaclient::transformItemName(const std::string& s)
{
    return ireplace_first_copy(s, "prd_", "");
}
