#ifndef rfaclientRFAPP_UTILITY_HPP
#define rfaclientRFAPP_UTILITY_HPP

#include "types.hpp"

/* RFA includes */
#include <Common/Context.h>
#include <Common/Event.h>
#include <Common/EventQueue.h>
#include <Common/EventQueueGroup.h>
#include <Common/TokenizedPrincipalIdentity.h>
#include <Config/ConfigDatabase.h>
#include <Config/ConfigTree.h>
#include <Config/StagingConfigDatabase.h>
#include <Logger/AppLogger.h>
#include <Logger/AppLoggerMonitor.h>

#include <SessionLayer/MarketDataContributor.h>
#include <SessionLayer/MarketDataItemEvent.h>
#include <SessionLayer/MarketDataItemContReplyStatus.h>
#include <SessionLayer/MarketDataSubscriber.h>
#include <SessionLayer/Session.h>

#include <TIBMsg/TibMsg.h>

/* Boost includes */
#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>

#include <stdexcept>
#include <iostream>

namespace rfaclient {

	typedef std::runtime_error MarketDataError;

	/*
	* Wrapper for the RFA global Context object
	*/
    struct  AutoContext : private boost::noncopyable {
        ~AutoContext() 
        {
			rfa::common::Context::uninitialize();
        }
        AutoContext()
        {
            if (!rfa::common::Context::initialize())
				throw MarketDataError("Context::initialize() failed");
        }
    };

    template <typename T>
    void
    destroy(T* p)
    {
        p->destroy();
    }

    template <typename T>
    void
    release(T* p)
    {
        p->release();
    }

    template <typename T>
    boost::shared_ptr<T>
    makeCreatePtr(T* p)
    {
        if (!p)
            throw MarketDataError("Failed to create RFA resource");
        return boost::shared_ptr<T>(p, destroy<T>);
    }

    template <typename T>
    boost::shared_ptr<T>
    makeAcquirePtr(T* p)
    {
        if (!p)
            throw MarketDataError("Failed to acquire RFA resource");
        return boost::shared_ptr<T>(p, release<T>);
    }

    //! Returns true if the field contains a value, i.e. it is supported,
    //! not data and not null.

     bool
    unpack(Field& dst, TibField& src);

    typedef boost::shared_ptr<
        rfa::config::StagingConfigDatabase> StagingConfigDatabasePtr;

    inline StagingConfigDatabasePtr
    createStagingConfigDatabase(const RFA_String& path)
	{
        StagingConfigDatabasePtr ptr
            (makeCreatePtr(rfa::config::StagingConfigDatabase::create()));

        if (!ptr->load(rfa::config::flatFile, path))
            throw MarketDataError("StagingConfigDatabase::load() failed");

        return ptr;
    }

    typedef boost::shared_ptr<rfa::config::ConfigDatabase> ConfigDatabasePtr;

    inline ConfigDatabasePtr
    createConfigDatabase(const rfa::config::StagingConfigDatabase& stagingDatabase)
    {
		ConfigDatabasePtr ptr
            (makeAcquirePtr(rfa::config::ConfigDatabase::acquire(rfa::common::Context::getName())));
        if (!ptr->merge(stagingDatabase))
            throw MarketDataError("ConfigDatabase::merge() failed");

        return ptr;
    }

    typedef boost::shared_ptr<
        rfa::logger::ApplicationLogger> ApplicationLoggerPtr;

    inline ApplicationLoggerPtr
    createApplicationLogger()
    {
        return ApplicationLoggerPtr
            (makeAcquirePtr(rfa::logger::ApplicationLogger
                            ::acquire(rfa::common::Context::getName())));
    }

    typedef boost::shared_ptr<
        rfa::logger::ComponentLogger> ComponentLoggerPtr;

    inline ComponentLoggerPtr
    createComponentLogger(rfa::logger::ApplicationLogger& applicationLogger,
                          const RFA_String& name)
    {
        return ComponentLoggerPtr
            (makeCreatePtr(applicationLogger.createComponentLogger(name)));
    }

    typedef boost::shared_ptr<rfa::common::EventQueue> EventQueuePtr;

    inline EventQueuePtr
    createEventQueue(const RFA_String& name)
    {
        return EventQueuePtr(makeCreatePtr(rfa::common::EventQueue
                                           ::create(name)));
    }

    typedef boost::shared_ptr<
        rfa::common::EventQueueGroup> EventQueueGroupPtr;

    inline EventQueueGroupPtr
    createEventQueueGroup(const RFA_String& name)
    {
        return EventQueueGroupPtr(makeCreatePtr(rfa::common::EventQueueGroup
                                                ::create(name)));
    }

    typedef boost::shared_ptr<
        rfa::logger::AppLoggerMonitor> AppLoggerMonitorPtr;

    inline AppLoggerMonitorPtr
    createAppLoggerMonitor(rfa::logger::ApplicationLogger& applicationLogger,
                           const RFA_String& name)
    {
        return AppLoggerMonitorPtr
            (makeCreatePtr(applicationLogger
                           .createApplicationLoggerMonitor(name, false)));
    }

    typedef boost::shared_ptr<rfa::sessionLayer::Session> SessionPtr;

    inline SessionPtr
    createSession(const RFA_String& name)
    {
        return SessionPtr(makeAcquirePtr(rfa::sessionLayer::Session
                                         ::acquire(name)));
    }

	inline SessionPtr
		 createSession(const RFA_String& name, const PrincipalIdentity* identity)
    {
        return SessionPtr(makeAcquirePtr(rfa::sessionLayer::Session
                                         ::acquire(name, identity)));
    }

    typedef boost::shared_ptr<
        rfa::sessionLayer::MarketDataSubscriber> MarketDataSubscriberPtr;

    inline MarketDataSubscriberPtr
    createMarketDataSubscriber(rfa::sessionLayer::Session& session,
                               const RFA_String& name)
    {
        return MarketDataSubscriberPtr
            (makeCreatePtr(session.createMarketDataSubscriber(name)));
    }

    typedef boost::shared_ptr<
        rfa::sessionLayer::MarketDataContributor> MarketDataContributorPtr;

    inline MarketDataContributorPtr
    createMarketDataContributor(rfa::sessionLayer::Session& session,
                                const RFA_String& name)
    {
        return MarketDataContributorPtr
            (makeCreatePtr(session.createMarketDataContributor(name)));
    }

    inline void
    readMfeedDictionary(const char *appendix, const char *enumType)
    {
        TibMsg_error retCode;
		if ((retCode = TibMsg::ReadMfeedDictionary(appendix, enumType)) != TIBMSG_OK) {
            throw MarketDataError("TibMsg::ReadMfeedDictionary() failed");
		}
    }

    typedef rfa::sessionLayer::MarketDataItemEvent
    ::MarketDataMessageType MarketDataMessageType;
    typedef rfa::sessionLayer::MarketDataItemStatus::State State;
    typedef rfa::sessionLayer::MarketDataItemStatus::StatusCode StatusCode;

    typedef rfa::sessionLayer::MarketDataItemContReplyStatus
    ::State ContReplyState;
    typedef rfa::sessionLayer::MarketDataItemContReplyStatus
    ::StatusCode ContReplyStatusCode;

     const char*
    toString(MarketDataMessageType msgType);

     std::string
    toString(const rfa::common::Exception& e);
	
     std::pair<std::string, std::string>
    splitPair(const std::string& s, const char* delims = ":=");

     std::string
    transformItemName(const std::string& s);
}

#endif // rfaclientRFAPP_UTILITY_HPP
