//
// This file contains the implementation of TSClientByCount.
//

#include "rtr/rtrnotif.h"		// Access to event loop
#include "tsbycount.h"

#include <sstream>

//
// Constructor
//
TSClientByCount::TSClientByCount(
								 RTRTimeSeriesService& service,
								 const char *symbol,
								 int numSamples,
								 RTRTimeSeriesService::Period period
								 )
								 : _series(service.timeSeries(symbol, period, *this)), 
								 _numSamples(numSamples), _byDate(false)
{
	// We must wait until the tsValDefDb is complete locally  before we make the time
	// request
	//const_cast<RTRTSValDefDb&> (service.tsValDefDb()).addClient(*this);
	((RTRTSValDefDb&) (service.tsValDefDb())).addClient(*this);
}

TSClientByCount::TSClientByCount(
								 RTRTimeSeriesService& service,
								 const char *symbol,
								 const char* startDate,
								 const char* endDate,
								 RTRTimeSeriesService::Period period
								 )
								 : _series(service.timeSeries(symbol, period, *this)), 
								 _byDate(true)
{
	// We must wait until the tsValDefDb is complete locally  before we make the time
	// request
	//const_cast<RTRTSValDefDb&> (service.tsValDefDb()).addClient(*this);
	((RTRTSValDefDb&) (service.tsValDefDb())).addClient(*this);

	_startDate = new RTRDateTime(startDate, RTRTRUE);
	_endDate= new RTRDateTime(endDate, RTRTRUE);
}


//
// Event processing
//
void TSClientByCount::processSeriesComplete(RTRTimeSeries& series) 
{ 
	int initialized = 0;

	if (series.count() > 0) {
		SEXP names;
		int cols = 0, rows = 0;
		int rowIndex = 0;

		// Preprocessing step : calculate the number of valid samples
		for (int i = series.lower(); i <= series.upper(); ++i) {
			RTRTimeSample& sample = series[i];
			if(sample.valid())
				rows++;	
		}


		for (int i = series.lower(); i <= series.upper(); ++i) 
		{

			// Initialize the data frame columns
			if (!initialized && series[series.lower()].valid()) 
			{
				RTRTimeSample& sample = series[series.lower()];

				cols = sample.upper() + 2; // extra column for 'Date' field
				//rows = series.upper() + 1;
				PROTECT(frame = allocVector(VECSXP, cols));
				PROTECT(names = allocVector(STRSXP, cols));

				SEXP dateColumn;
				PROTECT(dateColumn = allocVector(STRSXP, rows));
				SET_VECTOR_ELT(frame, 0, dateColumn);
				SET_STRING_ELT(names, 0, mkChar("Date"));

				for (int j = sample.lower(); j <= sample.upper(); ++j) 
				{
					int colIndex = j + 1;
					SEXP col;

					if (sample[j].dataType() < 2) 
						PROTECT(col = allocVector(REALSXP, rows));
					else
						PROTECT(col = allocVector(STRSXP, rows));

					SET_VECTOR_ELT(frame, colIndex, col);
					SET_STRING_ELT(names, colIndex, mkChar(sample[j].name())); 
				}

				initialized = 1;
			}
			
			RTRTimeSample& sample = series[i];

			if (sample.valid()) {
				stringstream ss;
				const RTRDateTime& dateTime = sample.dateTime();
				ss << dateTime.dayOfMonth() << "/" << dateTime.monthNumber() << "/" << dateTime.year();
				SET_STRING_ELT(VECTOR_ELT(frame, 0), rowIndex, mkChar(ss.str().c_str()));
				for (int j = sample.lower(); j <= sample.upper(); ++j) 
				{
					int colIndex = j+1;
					RTRSeriesValue& value = sample[j];
					
					if (value.dataType() < 2) 
					{ 
						double* vec =  REAL(VECTOR_ELT(frame, colIndex));
						vec[rowIndex] = (double)value;
					}
					else 
					{
						SET_STRING_ELT(VECTOR_ELT(frame, colIndex), rowIndex, mkChar(value));
					}			
				}

				rowIndex++;
			}
		}

		setAttrib(frame, R_NamesSymbol, names);
		UNPROTECT(cols + 2);
	}
	else 
		dataReturned = false;

	RTREventNotifierInit::notifier->disable();
}

SEXP& TSClientByCount::getFrame() {
	return frame;
}

void TSClientByCount::processSeriesError(RTRTimeSeries& series)
{
	cerr << series.symbol() << "Error :" << series.text() << endl;
	RTREventNotifierInit::notifier->disable();
}

void TSClientByCount::processSeriesInfo(RTRTimeSeries& s)
{
	//cerr << s.symbol() << s.text() << endl;
}


void TSClientByCount::processDefDbUpdate(RTRTSValDefDb& db)
{ //nothing to be done untill the defdb is 
}


void TSClientByCount::processDefDbComplete(RTRTSValDefDb& db)
{

	if (!_byDate)
		_series.setView(_numSamples);
	else {
		_series.setView(*_startDate, *_endDate);
	}

	if (!_series.error())
	{
		if (_series.complete())
			processSeriesComplete(_series);
		else
			processSeriesInfo(_series);
	}
	else
		processSeriesError(_series);
}
